clear all;
close all;
addpath('cartpole/');
settings_cp; % Sets parameters for the simulation, start states etc

%% SETUP
% Cartpole state:
%       p:        [m]     position of cart
%       dp:       [m/s]   velocity of cart
%       dtheta:   [rad/s] angular velocity
%       theta:    [rad]   angle
% Control:
%       u:        [N]     force on cart
% Dimensions of state and control
sdim = 4; udim = 1;

% Dynamics model input:
%       [p,dp,dtheta,sin(theta),cos(theta),u]
% Dynamics prediction: (separate GP for each output)
%       [dp,d^2p,d^2theta,dtheta] % delta-state, not next state
dyni = 6; dyno = 4;

% GP parameters (one per GP) (Try the effect of different parameters)
kernelLengthScale = [203.0197  242.9594  240.5070  218.0256
                     33.0219  176.8396  175.9314  178.0185
                     34.6307    7.3903    7.4687   13.0914
                      2.3903    1.0499    0.8433    1.2963
                     31.2894    0.9858    0.7810    1.7216
                    219.1850   24.6355   23.1603   49.9782]; % separate length for each dimension
kernelScaleFactor = [1.1478    1.3658    3.5236    0.7204]; % \sigma_f
noiseSigma        = [0.0143    0.0165    0.0431    0.0145]; % \sigma_n

% Init
numTrainEpochs = 25; 
N     = 10; 

% Generate an initial rollout on the cartpole using full dynamics
% xhat = [p_t, dp_t, dtheta_t, theta_t, sin(theta_t), cos(theta_t), u_t];
% x_next = [p_(t+1), dp_(t+1), dtheta_(t+1), theta_(t+1)] (dynamics predictions for training data)
[xhat_s, x_next_s] = rollout(gaussian(mu0, S0), policy, H, plant);
x_s  = xhat_s(:, plant.odei)'; % odei = state variable indices
U    = xhat_s(:, end-udim+1:end)'; % controls

% Use this rollout for initial training data
% NOTE: WE are predicting delta-state, so we use: state_(t+1) - state_t
trainX = xhat_s'; trainX(plant.angi,:) = []; % Remove the angle co-ordinate
trainY = (x_next_s - xhat_s(:,plant.odei))'; 

% Run the training loop. Use the full dynamics (cartpole_dynamics.m) to 
% generate the training dataset at each epoch. 
% Compare known dynamics against predictions from the GP model
for k = 1:numTrainEpochs
  
  % Switch between random and learned policy
  if (~mod(k,4)) % every 4th iteration
    disp('Using learned policy');
    policy = dataset.policy; % Use previously learned policy
    S0 = diag([0.05, 0.01, 0.01, 0.05].^2); % Change initial state distribution
  else
    disp('Using random policy');
    policy = []; policy.maxU = 10; % random policy
    S0 = diag([0.1 0.01 0.01 pi].^2);   % initial state covariance
  end
  
  % Generate a initial state by sampling from a gaussian
  x0 = gaussian(mu0, S0);

  %% EVALUATE THE GP MODEL AND FULL DYNAMICS USING THE CONTROL TRAJECTORY
  % 1) Get the result from the full dynamics
  % THIS WILL BE USED AS THE TRAINING DATA AT THE END OF THE CURRENT EPOCH
  [xhat_s, x_next_s] = rollout(x0, policy, H, plant);
  x_s = xhat_s(:, plant.odei)'; % odei = state variable indices
  U   = xhat_s(:, end-udim+1:end)'; % controls
  
  %% STUDENT TODO:
  % 2a) TODO: Get the result from the GP model
  % Go sequentially along the trajectory
  % Condition on the training data [trainX, trainY] to generate predictions and variances
  % You need to have one GP per output
  % All the 4 variables below need to be populated
  dx   	 = zeros(dyno, H);
  dx_var = ones(dyno, H);
  x 	 = zeros(dyno, H); 
  x_var  = ones(dyno, H);
  
  % 2b) TODO: Sample "N" trajectories using the previously computed means 
  % and variances. These will be used to display how the uncertainty
  % accumulates as we go through a long-term rollout over many steps
  %
  % All trajectories start at z0. 
  % At each prediction step, instead of using the mean-prediction, sample
  % from gaussian around the mean-prediction using the computed variance.
  % Use this sample for the next prediction and repeat.
  % The control trajectory is the same for all trajectories.
  % (z^j_0,u_0) ->  \mu^j_0 & \sigma^j_0 ->
  % s= gaussian(\mu^j_0,  \sigma^j_0) -> (z^j_0 + s, u_1) -> .....
  %
  % The sampling from the gaussian is the only extra step compared to 2a)
  % All the 4 variables below need to be populated
  dx_j 		= zeros(dyno, H, N);
  dx_j_var  = ones(dyno, H, N);
  x_j  		= zeros(dyno, H, N); 
  x_j_var   = ones(dyno, H, N);
  
  %% END TODO
  % 3) Add the full dynamics predictions to the training data
  % NOTE: WE are predicting delta-state, so we use: state_(t+1) - state_t
  augX = xhat_s'; augX(plant.angi,:) = []; % Remove the angle co-ordinate
  augY = (x_next_s - xhat_s(:,plant.odei))';
  trainX = [trainX augX];
  trainY = [trainY augY];
    
  % 4) Display the predictions from the full dynamics and GP side-by-side
  figure(fg);
  for j= 1:H
    
    % Display full dynamics result
    subplot(pltHdls(1)); cla;
    title('Full-Dynamics prediction','fontweight','bold','FontName','Times New Roman','fontsize',14);
    draw_cp(x_s(:,j), U(:,j), ['T=' num2str(j*dt) 'sec']);
    
    % Display GP result
    subplot(pltHdls(2)); cla;
    title(sprintf('GP-Dynamics prediction \n Training epoch: %d, Num Training data: %d',k,size(trainX,2)-H), ...
      'fontweight','bold','FontName','Times New Roman','fontsize',14);
    draw_cp(x(:,j), U(:,j), ['T=' num2str(j*dt) 'sec'], squeeze(x_j(:,j,:)));
    
    % Show plots of GP prediction
    for i = 1:dyno
      subplot(pltHdls(i+2)); cla; hold on;
      deltaTs = dt:dt:dt*j;
      plot(deltaTs, augY(i,1:j),'g-','Linewidth',2.5);
      plot(deltaTs, dx(i,1:j),'r-','Linewidth',2.5);
      ebar = errorbar(deltaTs, dx(i,1:j), 3*sqrt(dx_var(i,1:j)));
      set(ebar, 'Color', 'r','Linewidth',1.5);
      xlabel('Time (s)','fontweight','bold','fontsize',14,'FontName','Times New Roman');
      ylabel(predTxt{i},'fontweight','bold','fontsize',14,'FontName','Times New Roman');
      xlim([0,totT]);
    end
    
    % Pause to display
    drawnow;
    pause(0.1*dt);
  end
  
  % 5) Compare the mean-predictions from the GP and the full dynamics
  % Leave out the starting point
  x(4,:) = mod(x(4,:),2*pi);
  x_s(4,:) = mod(x_s(4,:),2*pi);
  sigmaStarsSquared = bsxfun(@plus, x_var, (noiseSigma').^2);
  SLL = 0.5 * log(2 * pi * sigmaStarsSquared) + (x_s - x).^2 ./ (2 * sigmaStarsSquared);
  disp(['Epoch: ', num2str(k), '. Mean Standardized Log Loss for each GP: ']);
  disp(num2str(mean(SLL,2)'));
  
end
